import json
from io import BytesIO
from pathlib import Path
from unittest.mock import call, patch

from arkindex_worker.models import Element
from worker_file_import.iiif import CollectionParser

SAMPLES = Path(__file__).absolute().parent / "samples"


def test_create_folder_kwargs(mock_worker):
    mock_worker.process_information = {
        "id": "processid",
        "mode": "files",
        "element_type": "page",
        "folder_type": "thing",
        "corpus": "corpusid",
    }
    mock_worker.api_client.add_response(
        "CreateElement",
        body={
            "corpus": "corpusid",
            "type": "thing",
            "name": "foldername",
            "worker_run_id": mock_worker.worker_run_id,
        },
        response={"id": "folderid", "name": "foldername", "type": "thing"},
    )
    parser = CollectionParser(
        mock_worker,
        BytesIO(b"{}"),
        first_level=False,
        folder_name="foldername",
    )
    parser.create_folder()
    assert "parent_element" in parser.parser_kwargs
    assert parser.parser_kwargs["parent_element"] == Element(
        {"id": "folderid", "name": "foldername", "type": "thing"}
    )


def test_create_folder_first_level(mock_worker):
    parser = CollectionParser(
        mock_worker,
        BytesIO(),
        first_level=True,
        corpus_id="corpusid",
        folder_type="thing",
    )
    parser.create_folder()
    assert parser.parser_kwargs.get("parent_id") is None


@patch("worker_file_import.iiif.IIIFParser.load")
def test_run_from(parser_mock, mock_worker, responses):
    responses.add(
        responses.GET,
        "http://iiif/subcollection/",
        json={"@id": "http://iiif/subcollection/"},
    )
    responses.add(
        responses.GET, "http://iiif/manifest/", json={"@id": "http://iiif/manifest/"}
    )

    with (SAMPLES / "collection.json").open("rb") as f:
        parser = CollectionParser(mock_worker, f)
        assert parser._run_from("members") == 2

    responses.assert_call_count("http://iiif/subcollection/", 1)
    responses.assert_call_count("http://iiif/manifest/", 1)


@patch("worker_file_import.iiif.logger")
@patch("worker_file_import.iiif.IIIFParser.load")
def test_run_from_exceptions(load_mock, logger_mock, mock_worker, responses):
    responses.add(responses.GET, "http://iiif/subcollection/", status=404)
    responses.add(
        responses.GET, "http://iiif/manifest/", json={"@id": "http://iiif/manifest/"}
    )
    load_mock().run.side_effect = ZeroDivisionError("Oh no")
    with (SAMPLES / "collection.json").open("rb") as f:
        parser = CollectionParser(mock_worker, f)
        assert parser._run_from("members") == 2

    responses.assert_call_count("http://iiif/subcollection/", 1)
    responses.assert_call_count("http://iiif/manifest/", 1)

    assert logger_mock.warning.call_count == 2
    assert logger_mock.warning.call_args_list == [
        call(
            "Failed fetching manifest/collection at http://iiif/subcollection/: 404 Client Error: Not Found for url: http://iiif/subcollection/",
            exc_info=False,
        ),
        call(
            "Error when running MagicMock for http://iiif/manifest/: Oh no",
            exc_info=False,
        ),
    ]


def test_run_from_not_found(mock_worker):
    parser = CollectionParser(mock_worker, BytesIO(b"{}"))
    assert parser._run_from("something") == 0


def test_run(mock_worker, responses):
    responses.add(
        responses.GET,
        "http://iiif/subcollection/",
        json={"@id": "http://iiif/subcollection/"},
    )
    responses.add(
        responses.GET, "http://iiif/manifest/", json={"@id": "http://iiif/manifest/"}
    )
    with (SAMPLES / "collection.json").open("rb") as f:
        parser = CollectionParser(mock_worker, f)
        parser.run()

    responses.assert_call_count("http://iiif/subcollection/", 1)
    responses.assert_call_count("http://iiif/manifest/", 1)


def test_run_fallback(mock_worker, responses):
    responses.add(
        responses.GET,
        "http://iiif/subcollection/fallback/",
        json={"@id": "http://iiif/subcollection/fallback/"},
    )
    responses.add(
        responses.GET,
        "http://iiif/manifest/fallback/",
        json={"@id": "http://iiif/manifest/fallback/"},
    )
    with (SAMPLES / "collection.json").open("rb") as f:
        data = json.load(f)

    # Remove `members` to force a fallback to `manifests` and `collections`
    del data["members"]
    parser = CollectionParser(mock_worker, BytesIO(json.dumps(data).encode("utf-8")))
    parser.run()

    responses.assert_call_count("http://iiif/manifest/fallback/", 1)
    responses.assert_call_count("http://iiif/subcollection/fallback/", 1)
