import json
from io import BytesIO
from pathlib import Path
from unittest.mock import patch

import pytest

from arkindex_worker.models import Element
from worker_file_import.iiif import ManifestParser

SAMPLES = Path(__file__).absolute().parent / "samples"


def test_create_folder_metadata(mock_worker):
    mock_worker.api_client.add_response(
        "CreateMetaDataBulk",
        body={
            "worker_run_id": mock_worker.worker_run_id,
            "metadata_list": [
                {
                    "name": "meta",
                    "type": "text",
                    "value": "data",
                    "entity_id": None,
                }
            ],
        },
        id="folderid",
        response={"metadata_list": [{"id": "mdid"}]},
    )
    parser = ManifestParser(
        mock_worker,
        stream=BytesIO(
            json.dumps(
                {
                    "@id": "something",
                    "metadata": [
                        {},
                        {"label": "nope"},
                        {"label": "nope", "value": ""},
                        {"value": "nope"},
                        {"label": "meta", "value": "data"},
                    ],
                }
            ).encode("utf-8")
        ),
    )
    parser.folder = {"id": "folderid", "name": "foldername"}
    parser.create_folder_metadata()


def test_create_metadata(mock_worker):
    mock_worker.api_client.add_response(
        "CreateMetaDataBulk",
        body={
            "worker_run_id": mock_worker.worker_run_id,
            "metadata_list": [
                {
                    "name": "some label",
                    "type": "text",
                    "value": "some value",
                    "entity_id": None,
                }
            ],
        },
        id="elementid",
        response={"metadata_list": [{"id": "mdid"}]},
    )
    parser = ManifestParser(mock_worker, BytesIO())
    parser.create_metadata(
        element={"id": "elementid", "name": "An element"},
        metadatas=[{"label": "some label", "value": "some value"}],
    )


def test_create_metadata_url(mock_worker):
    mock_worker.api_client.add_response(
        "CreateMetaDataBulk",
        body={
            "worker_run_id": mock_worker.worker_run_id,
            "metadata_list": [
                {
                    "name": "not a url",
                    "type": "text",
                    "value": "gopher://arkindex.teklia.com:70/1/",
                    "entity_id": None,
                },
                {
                    "name": "a url",
                    "type": "url",
                    "value": "https://arkindex.teklia.com/",
                    "entity_id": None,
                },
            ],
        },
        id="elementid",
        response={"metadata_list": [{"id": "mdid"}, {"id": "mdid2"}]},
    )
    parser = ManifestParser(mock_worker, BytesIO())
    parser.create_metadata(
        element={"id": "elementid", "name": "An element"},
        metadatas=[
            {"label": "not a url", "value": "gopher://arkindex.teklia.com:70/1/"},
            {"label": "a url", "value": "https://arkindex.teklia.com/"},
        ],
    )


def test_get_image(mock_worker, responses):
    responses.add(
        responses.GET, "http://imageurl/info.json", json={"@id": "http://imageurl/"}
    )
    mock_worker.api_client.add_response(
        "CreateIIIFInformation",
        body={"@id": "http://imageurl/"},
        response={"id": "imageid", "url": "http://imageurl", "status": "checked"},
    )

    parser = ManifestParser(mock_worker, BytesIO())
    assert parser.get_image("http://imageurl") == {
        "id": "imageid",
        "url": "http://imageurl",
        "status": "checked",
    }
    responses.assert_call_count("http://imageurl/info.json", 1)


def test_get_image_check_error(mock_worker, responses):
    responses.add(
        responses.GET, "http://imageurl/info.json", json={"@id": "http://imageurl/"}
    )
    mock_worker.api_client.add_error_response(
        "CreateIIIFInformation",
        body={"@id": "http://imageurl/"},
        content={"status": ["Image check failed!!1!"], "id": "imageid"},
        status_code=400,
    )
    mock_worker.api_client.add_response(
        "RetrieveImage",
        id="imageid",
        response={"id": "imageid", "url": "http://imageurl/", "status": "error"},
    )
    parser = ManifestParser(mock_worker, BytesIO())
    assert parser.get_image("http://imageurl/") == {
        "id": "imageid",
        "url": "http://imageurl/",
        "status": "error",
    }
    responses.assert_call_count("http://imageurl/info.json", 1)


@patch(
    "worker_file_import.utils.retried_get.retry.wait.wait_fixed",
    new=0,
)
def test_get_image_epic_fail(mock_worker, responses):
    responses.add(responses.GET, "http://imageurl/info.json", status=500)
    parser = ManifestParser(mock_worker, BytesIO())
    assert parser.get_image("http://imageurl") is None
    responses.assert_call_count("http://imageurl/info.json", 2)


def test_parse_canvas(mock_worker, responses):
    mock_worker.process_information = {
        "id": "processid",
        "mode": "files",
        "element_type": "thing",
    }
    responses.add(
        responses.GET, "http://imageurl/info.json", json={"@id": "http://imageurl/"}
    )
    mock_worker.api_client.add_response(
        "CreateIIIFInformation",
        body={"@id": "http://imageurl/"},
        response={
            "id": "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa",
            "url": "http://imageurl",
            "status": "checked",
        },
    )
    mock_worker.api_client.add_response(
        "CreateElement",
        body={
            "corpus": "corpusid",
            "parent": "folderid",
            "type": "thing",
            "name": "Thing 1",
            "image": "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa",
            "worker_run_id": mock_worker.worker_run_id,
            "confidence": None,
            "polygon": None,
        },
        response={"id": "elementid", "name": "Thing 1", "type": "thing"},
    )
    parser = ManifestParser(mock_worker, BytesIO(), corpus_id="corpusid")
    parser.folder = {
        "id": "folderid",
        "name": "foldername",
        "corpus": {"id": "corpusid"},
    }

    parser.parse_canvas(
        {
            "label": "Thing 1",
            "images": [{"resource": {"service": {"@id": "http://imageurl"}}}],
        }
    )

    assert parser.elements == [{"id": "elementid", "type": "thing"}]
    responses.assert_call_count("http://imageurl/info.json", 1)


def test_parse_canvas_no_name(mock_worker):
    parser = ManifestParser(mock_worker, BytesIO())
    with pytest.raises(ValueError, match="Canvas has no name"):
        parser.parse_canvas({})


def test_parse_canvas_no_image(mock_worker):
    parser = ManifestParser(mock_worker, BytesIO())
    with pytest.raises(ValueError, match="Canvas My canvas has no image"):
        parser.parse_canvas({"label": "My canvas", "images": []})


def test_parse_canvas_no_resource(mock_worker):
    parser = ManifestParser(mock_worker, BytesIO())
    with pytest.raises(ValueError, match="Canvas My canvas image has no resource"):
        parser.parse_canvas({"label": "My canvas", "images": [{}]})


def test_parse_canvas_no_service(mock_worker):
    parser = ManifestParser(mock_worker, BytesIO())
    with pytest.raises(
        ValueError, match="Image resource A has no IIIF image information service"
    ):
        parser.parse_canvas(
            {"label": "My canvas", "images": [{"resource": {"@id": "A"}}]}
        )


def test_parse_canvas_no_service_id(mock_worker):
    parser = ManifestParser(mock_worker, BytesIO())
    with pytest.raises(ValueError, match="Image service on resource A has no @id"):
        parser.parse_canvas(
            {
                "label": "My canvas",
                "images": [{"resource": {"@id": "A", "service": {}}}],
            }
        )


@patch(
    "worker_file_import.utils.retried_get.retry.wait.wait_fixed",
    new=0,
)
def test_parse_canvas_image_error(mock_worker, responses):
    mock_worker.process_information = {
        "id": "processid",
        "mode": "files",
        "element_type": "thing",
    }
    responses.add(responses.GET, "http://imageurl/info.json", status=500)
    mock_worker.api_client.add_response(
        "CreateElement",
        body={
            "corpus": "corpusid",
            "parent": "folderid",
            "type": "thing",
            "name": "Thing 1",
            "image": None,
            "worker_run_id": mock_worker.worker_run_id,
            "confidence": None,
            "polygon": None,
        },
        response={"id": "elementid", "name": "Thing 1", "type": "thing"},
    )
    parser = ManifestParser(mock_worker, BytesIO(), corpus_id="corpusid")
    parser.folder = {
        "id": "folderid",
        "name": "foldername",
        "corpus": {"id": "corpusid"},
    }

    parser.parse_canvas(
        {
            "label": "Thing 1",
            "images": [{"resource": {"service": {"@id": "http://imageurl"}}}],
        }
    )

    assert parser.elements == [{"id": "elementid", "type": "thing"}]
    responses.assert_call_count("http://imageurl/info.json", 2)


def test_parse_canvases(mock_worker, responses):
    mock_worker.process_information = {
        "id": "processid",
        "mode": "files",
        "element_type": "thing",
        "corpus": "corpusid",
        "folder_type": "folder",
    }
    responses.add(
        responses.GET,
        "http://iiif/image2/info.json",
        json={"@id": "http://iiif/image2/"},
    )
    responses.add(
        responses.GET,
        "http://iiif/image1/info.json",
        json={"@id": "http://iiif/image1/"},
    )

    mock_worker.api_client.add_response(
        "CreateIIIFInformation",
        body={"@id": "http://iiif/image2/"},
        response={
            "id": "aaaaaaaa-2222-aaaa-aaaa-aaaaaaaaaaaa",
            "url": "http://iiif/image2/",
            "status": "checked",
        },
    )
    mock_worker.api_client.add_response(
        "CreateIIIFInformation",
        body={"@id": "http://iiif/image1/"},
        response={
            "id": "aaaaaaaa-1111-aaaa-aaaa-aaaaaaaaaaaa",
            "url": "http://iiif/image1/",
            "status": "checked",
        },
    )
    mock_worker.api_client.add_response(
        "CreateElement",
        body={
            "corpus": "corpusid",
            "parent": "folderid",
            "type": "thing",
            "name": "1",
            "image": "aaaaaaaa-2222-aaaa-aaaa-aaaaaaaaaaaa",
            "worker_run_id": mock_worker.worker_run_id,
            "confidence": None,
            "polygon": None,
        },
        response={"id": "element1", "name": "1", "type": "thing"},
    )
    mock_worker.api_client.add_response(
        "CreateElement",
        body={
            "corpus": "corpusid",
            "parent": "folderid",
            "type": "thing",
            "name": "Canvas 1",
            "image": "aaaaaaaa-1111-aaaa-aaaa-aaaaaaaaaaaa",
            "worker_run_id": mock_worker.worker_run_id,
            "confidence": None,
            "polygon": None,
        },
        response={"id": "element2", "name": "Canvas 1", "type": "thing"},
    )
    mock_worker.api_client.add_response(
        "CreateMetaDataBulk",
        body={
            "worker_run_id": mock_worker.worker_run_id,
            "metadata_list": [
                {"type": "text", "name": "Méta", "value": "Donnée", "entity_id": None}
            ],
        },
        id="element1",
        response={"metadata_list": [{"id": "mdid"}]},
    )

    parser = ManifestParser(mock_worker, stream=(SAMPLES / "manifest.json").open("rb"))
    parser.folder = {
        "id": "folderid",
        "name": "foldername",
        "corpus": {"id": "corpusid"},
    }
    try:
        parser.parse_canvases()
    finally:
        parser.close()

    assert parser.elements == [
        {"id": "element2", "type": "thing"},
        {"id": "element1", "type": "thing"},
    ]


def test_run(mock_worker, responses):
    mock_worker.process_information = {
        "id": "processid",
        "mode": "files",
        "element_type": "page",
        "corpus": "corpusid",
        "folder_type": "folder",
    }
    responses.add(
        responses.GET,
        "http://iiif/image2/info.json",
        json={"@id": "http://iiif/image2/"},
    )
    responses.add(
        responses.GET,
        "http://iiif/image1/info.json",
        json={"@id": "http://iiif/image1/"},
    )

    mock_worker.api_client.add_error_response(
        "CreateIIIFInformation",
        body={"@id": "http://iiif/image2/"},
        status_code=400,
        content={
            "status": ["Image check failed!!1!"],
            "id": "aaaaaaaa-2222-aaaa-aaaa-aaaaaaaaaaaa",
        },
    )
    mock_worker.api_client.add_response(
        "CreateIIIFInformation",
        body={"@id": "http://iiif/image1/"},
        response={
            "id": "aaaaaaaa-1111-aaaa-aaaa-aaaaaaaaaaaa",
            "url": "http://iiif/image1/",
            "status": "checked",
        },
    )
    mock_worker.api_client.add_response(
        "RetrieveImage",
        id="aaaaaaaa-2222-aaaa-aaaa-aaaaaaaaaaaa",
        response={
            "id": "aaaaaaaa-2222-aaaa-aaaa-aaaaaaaaaaaa",
            "url": "http://iiif/image2/",
            "status": "checked",
        },
    )

    mock_worker.api_client.add_response(
        "CreateElement",
        body={
            "corpus": "corpusid",
            "parent": "parentid",
            "type": "folder",
            "name": "The Manifest",
            "worker_run_id": mock_worker.worker_run_id,
        },
        response={
            "id": "folderid",
            "name": "The Manifest",
            "type": "folder",
            "corpus": {"id": "corpusid"},
        },
    )
    mock_worker.api_client.add_response(
        "CreateElement",
        body={
            "corpus": "corpusid",
            "parent": "folderid",
            "type": "page",
            "name": "1",
            "image": "aaaaaaaa-2222-aaaa-aaaa-aaaaaaaaaaaa",
            "worker_run_id": mock_worker.worker_run_id,
            "confidence": None,
            "polygon": None,
        },
        response={"id": "element1", "name": "1", "type": "page"},
    )
    mock_worker.api_client.add_response(
        "CreateElement",
        body={
            "corpus": "corpusid",
            "parent": "folderid",
            "type": "page",
            "name": "Canvas 1",
            "image": "aaaaaaaa-1111-aaaa-aaaa-aaaaaaaaaaaa",
            "worker_run_id": mock_worker.worker_run_id,
            "confidence": None,
            "polygon": None,
        },
        response={"id": "element2", "name": "2", "type": "page"},
    )

    mock_worker.api_client.add_response(
        "CreateMetaData",
        body={
            "type": "reference",
            "name": "IIIF ID",
            "value": "http://iiif/manifest/",
            "entity_id": None,
            "worker_run_id": mock_worker.worker_run_id,
        },
        id="folderid",
        response={"id": "mdid"},
    )
    mock_worker.api_client.add_response(
        "CreateMetaDataBulk",
        body={
            "worker_run_id": mock_worker.worker_run_id,
            "metadata_list": [
                {"type": "text", "name": "Meta", "value": "Data", "entity_id": None},
                {
                    "type": "text",
                    "name": "Origin",
                    "value": "Russia",
                    "entity_id": None,
                },
            ],
        },
        id="folderid",
        response={"metadata_list": [{"id": "mdid"}, {"id": "mdid"}]},
    )
    mock_worker.api_client.add_response(
        "CreateMetaDataBulk",
        body={
            "worker_run_id": mock_worker.worker_run_id,
            "metadata_list": [
                {"type": "text", "name": "Méta", "value": "Donnée", "entity_id": None}
            ],
        },
        id="element1",
        response={"metadata_list": [{"id": "mdid"}]},
    )

    parser = ManifestParser(
        mock_worker,
        stream=(SAMPLES / "manifest.json").open("rb"),
        parent_element=Element({"id": "parentid"}),
    )
    parser.run()

    assert parser.elements == [
        {"id": "folderid", "type": "folder"},
        {"id": "element2", "type": "page"},
        {"id": "element1", "type": "page"},
    ]
