import json
from io import BytesIO, StringIO
from pathlib import Path

import pytest

from arkindex_worker.models import Element
from worker_file_import.iiif import IIIFParser, ManifestParser

SAMPLES = Path(__file__).absolute().parent / "samples"


def test_stream_binary(mock_worker):
    """
    Check IIIFParser rejects streams opened in text mode
    """
    with pytest.raises(
        AssertionError, match="Stream should be a file-like object in binary mode"
    ):
        IIIFParser(mock_worker, StringIO("something"))


def test_close(mock_worker):
    stream = BytesIO()
    IIIFParser(mock_worker, stream).close()
    assert stream.closed


def test_v3(mock_worker):
    stream = BytesIO(
        json.dumps(
            {
                "@context": "http://iiif.io/api/presentation/3/context.json",
                "id": "something",
                "type": "sc:Manifest",
            }
        ).encode("utf-8")
    )
    with pytest.raises(
        ValueError,
        match="The import functionality does not support IIIF Presentation API version 3",
    ):
        IIIFParser(mock_worker, stream).check_version()


def test_get_type(mock_worker):
    stream = BytesIO(
        json.dumps({"@id": "something", "@type": "sc:Manifest"}).encode("utf-8")
    )
    assert IIIFParser(mock_worker, stream).get_type() == ManifestParser


def test_get_type_invalid(mock_worker):
    with pytest.raises(ValueError, match="Missing @type property in JSON data"):
        IIIFParser(mock_worker, BytesIO(b"{}")).get_type()


def test_get_type_unsupported(mock_worker):
    stream = BytesIO(
        json.dumps({"@id": "something", "@type": "sc:Fest"}).encode("utf-8")
    )
    with pytest.raises(ValueError, match='Unsupported @type "sc:Fest"'):
        IIIFParser(mock_worker, stream).get_type()


def test_load(mock_worker):
    stream = BytesIO(
        json.dumps(
            {
                "@id": "something",
                "@type": "sc:Manifest",
            }
        ).encode("utf-8")
    )
    assert isinstance(
        IIIFParser(worker=mock_worker, stream=stream).load(
            worker=mock_worker, stream=stream
        ),
        ManifestParser,
    )


def test_run(mock_worker):
    with pytest.raises(NotImplementedError):
        IIIFParser(mock_worker, BytesIO()).run()


def test_create_folder(mock_worker):
    mock_worker.process_information = {
        "id": "processid",
        "mode": "files",
        "element_type": "page",
        "folder_type": "folder",
        "corpus": "corpus_id",
    }
    mock_worker.api_client.add_response(
        "CreateElement",
        body={
            "corpus": "corpus_id",
            "type": "folder",
            "name": "Le Manifest",
            "worker_run_id": mock_worker.worker_run_id,
        },
        response={"id": "folderid", "name": "Le Manifest", "type": "folder"},
    )
    mock_worker.api_client.add_response(
        "CreateMetaData",
        body={
            "name": "IIIF ID",
            "type": "reference",
            "value": "something",
            "worker_run_id": mock_worker.worker_run_id,
            "entity_id": None,
        },
        id="folderid",
        response={"id": "mdid"},
    )

    parser = IIIFParser(
        mock_worker,
        stream=BytesIO(
            json.dumps(
                {"@id": "something", "@type": "sc:Manifest", "label": "Manifesto"}
            ).encode("utf-8")
        ),
        folder_name="Le Manifest",
    )
    parser.create_folder()

    assert parser.folder, {"id": "folderid", "name": "foldername", "type": "folder"}
    assert parser.elements == [{"id": "folderid", "type": "folder"}]


def test_create_folder_parent(mock_worker):
    mock_worker.process_information = {
        "id": "processid",
        "mode": "files",
        "element_type": "page",
        "folder_type": "volume",
        "corpus": "corpus_id",
    }
    mock_worker.api_client.add_response(
        "CreateElement",
        body={
            "corpus": "corpus_id",
            "parent": "parent_id",
            "type": "volume",
            "name": "Le Manifest",
            "worker_run_id": mock_worker.worker_run_id,
        },
        response={
            "id": "folderid",
            "name": "Le Manifest",
            "type": "volume",
            "parent": "parent_id",
        },
    )
    mock_worker.api_client.add_response(
        "CreateMetaData",
        body={
            "name": "IIIF ID",
            "type": "reference",
            "value": "something",
            "worker_run_id": mock_worker.worker_run_id,
            "entity_id": None,
        },
        id="folderid",
        response={"id": "mdid"},
    )

    parser = IIIFParser(
        mock_worker,
        stream=BytesIO(
            json.dumps(
                {"@id": "something", "@type": "sc:Manifest", "label": "Manifesto"}
            ).encode("utf-8")
        ),
        parent_element=Element({"id": "parent_id"}),
        folder_name="Le Manifest",
    )
    parser.create_folder()

    assert parser.folder, {"id": "folderid", "name": "foldername", "type": "volume"}
    assert parser.elements == [{"id": "folderid", "type": "volume"}]


def test_create_folder_label(mock_worker):
    mock_worker.process_information = {
        "id": "processid",
        "mode": "files",
        "element_type": "page",
        "folder_type": "volume",
        "corpus": "corpus_id",
    }
    mock_worker.api_client.add_response(
        "CreateElement",
        body={
            "corpus": "corpus_id",
            "type": "volume",
            "name": "Manifesto",
            "worker_run_id": mock_worker.worker_run_id,
        },
        response={"id": "folderid", "name": "Manifesto", "type": "volume"},
    )
    mock_worker.api_client.add_response(
        "CreateMetaData",
        body={
            "name": "IIIF ID",
            "type": "reference",
            "value": "something",
            "worker_run_id": mock_worker.worker_run_id,
            "entity_id": None,
        },
        id="folderid",
        response={"id": "mdid"},
    )

    parser = IIIFParser(
        mock_worker,
        stream=BytesIO(
            json.dumps(
                {"@id": "something", "@type": "sc:Manifest", "label": "Manifesto"}
            ).encode("utf-8")
        ),
    )
    parser.create_folder()

    assert parser.folder == {"id": "folderid", "name": "Manifesto", "type": "volume"}
    assert parser.elements == [{"id": "folderid", "type": "volume"}]


def test_create_folder_filename(mock_worker):
    mock_worker.process_information = {
        "id": "processid",
        "mode": "files",
        "element_type": "page",
        "folder_type": "folder",
        "corpus": "corpus_id",
    }
    mock_worker.api_client.add_response(
        "CreateElement",
        body={
            "corpus": "corpus_id",
            "type": "folder",
            "name": "el-manifesto",
            "worker_run_id": mock_worker.worker_run_id,
        },
        response={
            "id": "folderid",
            "corpus": "corpus_id",
            "type": "folder",
            "name": "el-manifesto",
        },
    )
    mock_worker.api_client.add_response(
        "CreateMetaData",
        body={
            "name": "IIIF ID",
            "type": "reference",
            "value": "something",
            "worker_run_id": mock_worker.worker_run_id,
            "entity_id": None,
        },
        id="folderid",
        response={"id": "mdid"},
    )

    stream = BytesIO(
        json.dumps({"@id": "something", "@type": "sc:Manifest"}).encode("utf-8")
    )
    stream.name = "/path/to/el-manifesto.json"

    parser = IIIFParser(mock_worker, stream=stream)
    parser.create_folder()

    assert parser.folder == {
        "id": "folderid",
        "name": "el-manifesto",
        "type": "folder",
        "corpus": "corpus_id",
    }
    assert parser.elements == [{"id": "folderid", "type": "folder"}]


def test_create_folder_no_name(mock_worker):
    parser = IIIFParser(
        mock_worker,
        stream=BytesIO(
            json.dumps({"@id": "something", "@type": "sc:Manifest"}).encode("utf-8")
        ),
        corpus_id="corpus_id",
        folder_type="volume",
    )
    with pytest.raises(AssertionError, match="Could not determine a folder name"):
        parser.create_folder()
